/* Md5 loader class.

   see md5.h for issues.

   this code is released under the GPL License

   (c) 2002 Sbastien Kuntz
   aka CbWan <cb@nowan.net>
*/

#include "md5.h"
#include "tgaload.h"

#include <GL/gl.h>

#include <math.h>
#include <cstdio>
#include <cstring>

#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <fstream>

using namespace std;

MD5::MD5( const char* meshfilename, const char* animfilename ):
  _nbMeshes(0),
  _meshes(0),
  _nbBones(0),
  _bones(0),
  _anim( false )
{
  if (readDOOMFile( meshfilename )) {
    if( !readANIMFile( animfilename ) ) {
      cout << "couldn't load animation : " << animfilename << endl;
    }
    else {
      _anim = true;
    }

    buildVertices();
    buildNormals();

    buildTextures();
  }
  else {
    cerr << "couldn't load mesh : " << meshfilename << endl;
  }
}

MD5::~MD5( void ) {
  for( int n_mesh = 0; n_mesh < _nbMeshes ; n_mesh++ ) {
    delete[]  _meshes[n_mesh].vertices;
    delete[]  _meshes[n_mesh].triangles;
    delete[]  _meshes[n_mesh].weights;   
    //XXX free textures
  }
}

void MD5::buildTextures( void ) {
  _textures = new GLuint[_nbMeshes];
  glGenTextures(_nbMeshes, &_textures[0]);

  for( int n_mesh=0; n_mesh < _nbMeshes; n_mesh++ ) {
    Mesh& mesh = _meshes[ n_mesh ];

    // dirk reiners was here ;)
    std::string name(strrchr(mesh.texture, '/') + 1); // just the file name

    bool found = true;

    // testing file existence
    FILE* file = fopen( name.c_str(), "rb" );

    if( !file ) {
      // Try to append _d to get the diffuse texture
      name.insert(name.rfind(".tga"),"_d");
      file = fopen( name.c_str(), "rb" );
      if( !file ) {
	name.replace(name.rfind("_d.tga"), 6, "_s.tga");
	file = fopen( name.c_str(), "rb" );
	if( !file ) {
	  found = false;
	  cout << "Warning : Couldn't find texture for " << mesh.texture << endl;
	}
      }
    }

    if( found ) {
      fclose(file);	    
      _textures[n_mesh] = tgaLoadAndBind( name.c_str(), TGA_FREE );
    }
  }   
}   

void MD5::setFrame( int frame )
{
  glPushMatrix();
  glLoadIdentity();
  recursiveBuildBones( frame, &_bones[0] );
  glPopMatrix();
  buildVertices();
  buildNormals();
}

void MD5::recursiveBuildBones( int frame, Bone* bone )
{
  /* in the md5anim, bones properties are in parent's
     coordinate system. I use the modelview matrix
     to get the absolute values needed by buildVertices */

 if( _anim ) {
  glPushMatrix();

  float trans[6];

  Channel* chan;
  int lframe;

  /* reading bone transform values from channels */
  for( int i=0; i<6; i++ ) {
    chan = bone->chan[i];
    lframe = frame%chan->nbKeys;
//    if( lframe >= chan->range[0] && lframe <= chan->range[1] )
      trans[i] = chan->keys[lframe];
//    else
//      trans[i] = chan->keys[chan->range[0]];
  }

  /* apply read values to the modelview matrix */
  glTranslatef( trans[X],trans[Y],trans[Z] );
  glRotatef(trans[YAW]  , 0,0,1 );
  glRotatef(trans[PITCH], 0,1,0 );
  glRotatef(trans[ROLL] , 1,0,0 );

  // saving current values
  GLfloat matrix[16];
  glGetFloatv( GL_MODELVIEW_MATRIX, matrix );

  bone->pos[0] = matrix[4*3 + 0 ];
  bone->pos[1] = matrix[4*3 + 1 ];
  bone->pos[2] = matrix[4*3 + 2 ];

  for(int i=0; i<3; i++) {
    for( int j=0; j<3; j++) {
      bone->rot[3*i+j] = matrix[4*i+j];
    }
  }
  
  vector<Bone*>::iterator it = bone->children.begin();

  for( ; it != bone->children.end(); it++ ) {
    recursiveBuildBones( frame, (*it));
  }

  glPopMatrix();
 }
}

Bone* MD5::findBone( const char* name ) {
  for( int i=0; i<_nbBones;i++ ) {
    if( !strcmp( name, _bones[i].name ) ) {
      return &_bones[i];
    }
  }
  if( string(name).find("boundsM") == string::npos )
    cout << "didn't find bone : " << name << endl;
  return 0;
}

bool MD5::readANIMFile( const char* filename )
{
  FILE* fp = fopen( filename, "rb" );
  if( fp == 0 )
    return false;

  // Skip header
  readLine(fp);
  readLine(fp);
  readLine(fp);
  
  // == CHANNELS ==
  readLine(fp);
  if(!memcmp(_currentLine,"numchannels",11)) {
    sscanf(_currentLine,"numchannels %d", &_nbChannels );
    _channels = new Channel[_nbChannels];
    
    printf("Number of channels: %d\n", _nbChannels);

    for( int i=0; i < _nbChannels; i++ ) {
      Channel& chan = _channels[i];

      // start of the channel
      readLine(fp);
      // channel i
      readLine(fp);

      // == JOINT ==
      readLine(fp);
      sscanf(_currentLine,"\tjoint \"%s",chan.joint);
      chan.joint[strlen(chan.joint) - 1] = 0;      

      chan.bone = findBone( chan.joint );

      // == ATTRIBUTE ==
      readLine(fp);
      
      char attr[10];
      sscanf(_currentLine,"\tattribute \"%s",attr);
      attr[strlen(attr)-1] = 0;

      string Attr(attr);

      if( Attr == "x" ) {
	chan.attribute = X;
      }
      else if( Attr == "y" ) {
	chan.attribute = Y;
      }
      else if( Attr == "z" ) {
	chan.attribute = Z;
      }
      else if( Attr == "yaw" ) {
	chan.attribute = YAW;
      }
      else if( Attr == "pitch" ) {
	chan.attribute = PITCH;
      }
      else if( Attr == "roll" ) {
	chan.attribute = ROLL;
      }

      // assign channel to bone
      if( chan.bone != 0 )
	chan.bone->chan[chan.attribute] = &chan;


      // == STARTTIME ==
      readLine(fp);
      sscanf(_currentLine,"\tstarttime %f",&chan.starttime);

      readLine(fp);
      sscanf(_currentLine,"\tendtime %f",&chan.endtime);

      readLine(fp);
      sscanf(_currentLine,"\tframerate %f",&chan.framerate);

      readLine(fp);
      readLine(fp);
      sscanf(_currentLine,"\tstrings %d",&chan.strings);

      readLine(fp);
      readLine(fp);
      sscanf(_currentLine,"\trange %d %d",&chan.range[0],&chan.range[1]);

      readLine(fp);
      sscanf(_currentLine,"\tkeys %d",&chan.nbKeys);

      chan.keys = new float[chan.nbKeys];

      int readKeys = 0;
      int readOnLine = 0;

      while( readKeys < chan.nbKeys ) {
	readLine(fp);
	istringstream keys( _currentLine );
	while( readOnLine < 8 ) {
	  keys >> chan.keys[readKeys++];
	  readOnLine++;
	  if( readKeys >= chan.nbKeys )
	    break;
	}
	readOnLine = 0;
      }
      readLine(fp);
    } 
  }
   else 
     cerr << "euh.. : " << _currentLine << endl;

  fclose(fp);
  return true;
}


/* beginning of adrian code (found on flipcode forum) */

bool MD5::readLine(FILE* fp) 
{
   int pos = 0;

   while (1)
   {
      int inchar = fgetc(fp);

      if (inchar == EOF)
      {
         _currentLine[pos] = 0;
         return (false);
      }

      if (inchar == 13)
      {
         fgetc(fp);
         break;
      }

      _currentLine[pos++] = inchar;
   }

   _currentLine[pos] = 0;

   return (true);
}

bool MD5::readDOOMFile(const char* filename)
{
   FILE* fp = fopen(filename, "rb");
   if (fp == 0)
      return (false);

   // Skip header
   readLine(fp);
   readLine(fp);

   // ========== BONES ==========
   readLine(fp);
   if (!memcmp(_currentLine, "numbones", 8))
   {
      sscanf(_currentLine, "numbones %d", &_nbBones);
      
      _bones = new Bone[ _nbBones ];      

      printf("Number of bones: %d\n", _nbBones);

      for (int i = 0; i < _nbBones; i++)
      {
         Bone& bone = _bones[i];

         // Start of the bone
         readLine(fp);

         // ========== NAME ==========
         readLine(fp);
         sscanf(_currentLine, "\tname \"%s", bone.name);
         bone.name[strlen(bone.name) - 1] = 0;

	 if( i==0 ) {
	   strcpy( bone.name, "origin" );
	   cout << "Creating BONE : " << bone.name << endl;
	 }

         // ========== POSITION ==========
         readLine(fp);
         sscanf(_currentLine, "\tbindpos %f %f %f", &bone.pos[0], &bone.pos[1], &bone.pos[2]);

         // ========== ROTATION ==========
         readLine(fp);
         sscanf(_currentLine, "\tbindmat %f %f %f %f %f %f %f %f %f",
            &bone.rot[0],
            &bone.rot[1],
            &bone.rot[2],
            &bone.rot[3],
            &bone.rot[4],
            &bone.rot[5],
            &bone.rot[6],
            &bone.rot[7],
            &bone.rot[8]);

         // ========== PARENT ==========
         readLine(fp);
         bone.parent = -1;
         if (_currentLine[0] != '}')
         {
            char parent[64];
            sscanf(_currentLine, "\tparent \"%s", parent);
            parent[strlen(parent) - 1] = 0;
            for (int j = 0; j < i; j++)
            {
               if (!strcmp(parent, _bones[j].name))
               {
                  bone.parent = j;		  
                  break;
               }
            }
            readLine(fp);
         }

         // Separating line
         readLine(fp);
      }
   }

   // ========== MESHES ==========
   readLine(fp);
   if (!memcmp(_currentLine, "nummeshes", 9))
   {
      sscanf(_currentLine, "nummeshes %d", &_nbMeshes);
      _meshes = new Mesh[_nbMeshes];

      printf("Number of meshes: %d\n", _nbMeshes);

      readLine(fp);

      for (int i = 0; i < _nbMeshes; i++)
      {
         printf("Mesh %d\n", i);

         Mesh& mesh = _meshes[i];

         // Start of mesh
         readLine(fp);

         // ========== SHADER ==========
         readLine(fp);
         char* ptr = _currentLine;
         while (*ptr != '\"')
            ptr++;
         strcpy(mesh.texture, ptr + 1);
         mesh.texture[strlen(mesh.texture) - 1] = 0;
         readLine(fp);
         printf("   Shader: %s\n", mesh.texture);

         // ========== VERTICES ==========
         readLine(fp);
         sscanf(_currentLine, "\tnumverts %d", &mesh.nb_vertices);
         printf("   Number of vertices: %d\n", mesh.nb_vertices);
         mesh.vertices = new Vertex[mesh.nb_vertices];
         for (int j = 0; j < mesh.nb_vertices; j++)
         {
            int index;
            Vertex& vertex = mesh.vertices[j];
            readLine(fp);
            sscanf(_currentLine, "\tvert %d %f %f %d %d",
               &index,
               &vertex.u,
               &vertex.v,
               &vertex.weight_index,
               &vertex.weight_count);
            vertex.v = 1 - vertex.v;
	    
         }
         readLine(fp);

         // ========== TRIANGLES ==========
         readLine(fp);
         sscanf(_currentLine, "\tnumtris %d", &mesh.nb_triangles);
         printf("   Number of triangles: %d\n", mesh.nb_triangles);
         mesh.triangles = new Triangle[mesh.nb_triangles];
         for (int j = 0; j < mesh.nb_triangles; j++)
         {
            int index;
            Triangle& triangle = mesh.triangles[j];
            readLine(fp);
            sscanf(_currentLine, "\ttri %d %d %d %d",
               &index,
               &triangle.a,
               &triangle.b,
               &triangle.c);
         }
         readLine(fp);

         // ========== WEIGHTS ==========
         readLine(fp);
         sscanf(_currentLine, "\tnumweights %d", &mesh.nb_weights);
         printf("   Number of weights: %d\n", mesh.nb_weights);
         mesh.weights = new Weight[mesh.nb_weights];
         for (int j = 0; j < mesh.nb_weights; j++)
         {
            int index;
            Weight& weight = mesh.weights[j];
            readLine(fp);
            sscanf(_currentLine, "\tweight %d %d %f %f %f %f",
               &index,
               &weight.bone,
               &weight.t,
               &weight.x,
               &weight.y,
               &weight.z);
         }
         readLine(fp);

         readLine(fp);
      }
   }

   fclose(fp);
   
   // et hop on build la list des children
   
   for( int nbone=0; nbone< _nbBones; nbone++ ) {
     if( _bones[nbone].parent != -1 ) {
       Bone& parent = _bones[ _bones[nbone].parent ];
       parent.children.push_back( &_bones[nbone] );
     }	 
   }
     
   return (true);
}



void MD5::buildVertices(void)
{
   for (int i = 0; i < _nbMeshes; i++)
   {
      Mesh& mesh = _meshes[i];

      for (int j = 0; j < mesh.nb_vertices; j++)
      {
         Vertex& v = mesh.vertices[j];

         v.x = 0.0f;
         v.y = 0.0f;
         v.z = 0.0f;

         for (int k = 0; k < v.weight_count; k++)
         {
            Weight& w = mesh.weights[v.weight_index + k];

            Bone& b = _bones[w.bone];

            v.x += (b.rot[0] * w.x + b.rot[3] * w.y + b.rot[6] * w.z + b.pos[0]) * w.t;
            v.z += (b.rot[1] * w.x + b.rot[4] * w.y + b.rot[7] * w.z + b.pos[1]) * w.t;
            v.y += (b.rot[2] * w.x + b.rot[5] * w.y + b.rot[8] * w.z + b.pos[2]) * w.t;
         }
      }
   }
}

void MD5::buildNormals(void)
{
   for (int i = 0; i < _nbMeshes; i++)
   {
      Mesh& mesh = _meshes[i];

      for (int j = 0; j < mesh.nb_vertices; j++)
      {
         mesh.vertices[j].nx = 0;
         mesh.vertices[j].ny = 0;
         mesh.vertices[j].nz = 0;
      }

      for (int j = 0; j < mesh.nb_triangles; j++)
      {
         Triangle& tri = mesh.triangles[j];

         Vertex& v0 = mesh.vertices[tri.a];
         Vertex& v1 = mesh.vertices[tri.b];
         Vertex& v2 = mesh.vertices[tri.c];

         float Ax = v1.x - v0.x;
         float Ay = v1.y - v0.y;
         float Az = v1.z - v0.z;

         float Bx = v2.x - v0.x;
         float By = v2.y - v0.y;
         float Bz = v2.z - v0.z;

         float nx = Ay * Bz - By * Az;
         float ny = -(Ax * Bz - Bx * Az);
         float nz = Ax * By - Bx * Ay;

         v0.nx += nx;
         v0.ny += ny;
         v0.nz += nz;

         v1.nx += nx;
         v1.ny += ny;
         v1.nz += nz;

         v2.nx += nx;
         v2.ny += ny;
         v2.nz += nz;
      }

      for (int j = 0; j < mesh.nb_vertices; j++)
      {
         Vertex& v = mesh.vertices[j];

         float len = sqrt(v.nx * v.nx + v.ny * v.ny + v.nz * v.nz);

         if (len)
         {
            v.nx /= len;
            v.ny /= len;
            v.nz /= len;
         }
      }
   }   
}
